//
// GameX (Release 2 - Vectors / Matricies / CameraX)
// Matrix Class Header
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
//

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <memory.h>
#include <math.h>

#ifndef MATRIX_DEF
	#define MATRIX_DEF
											// **** Switches	
	#define MATRIX_DEBUG					// Enable Debugging Output
	#define MATRIX_INITIALIZE				// Initializes vectors	

	#include "gamex-debug.hpp"
	#include "gamex-vector.hpp"

	#define MSG_GETROWS		27975
	#define MSG_GETCOLS		55781
	#define MSG_MULT		58154
	#define MSG_SIZE		27831
	#define MSG_SIZESAFE	65442
	#define MSG_INSERTROW	46865
	#define MSG_INSERTCOL	9135
	#define MSG_TRANSPOSE	51755
	#define MSG_IDENTITY	13042
	#define MSG_ROTATEX		5491
	#define MSG_ROTATEY		47272
	#define MSG_ROTATEZ		23517
	#define MSG_BASIS		58298
	#define MSG_DIV			60383
											// Forward Referencing
	class MatrixC;
	class MatrixI;
	class MatrixF;

	class Matrix {
	public:
		// Member Virtual Functions		
		virtual Matrix &operator= (unsigned char c)=0;
		virtual Matrix &operator= (int c)=0;
		virtual Matrix &operator= (double c)=0;		
		virtual Matrix &operator= (MatrixC &op)=0;
		virtual Matrix &operator= (MatrixI &op)=0;
		virtual Matrix &operator= (MatrixF &op)=0;
		
		virtual Matrix &operator+= (unsigned char c)=0;
		virtual Matrix &operator+= (int c)=0;
		virtual Matrix &operator+= (double c)=0;		
		virtual Matrix &operator+= (MatrixC &op)=0;
		virtual Matrix &operator+= (MatrixI &op)=0;
		virtual Matrix &operator+= (MatrixF &op)=0;

		virtual Matrix &operator-= (unsigned char c)=0;
		virtual Matrix &operator-= (int c)=0;
		virtual Matrix &operator-= (double c)=0;		
		virtual Matrix &operator-= (MatrixC &op)=0;
		virtual Matrix &operator-= (MatrixI &op)=0;
		virtual Matrix &operator-= (MatrixF &op)=0;

		virtual Matrix &operator*= (unsigned char c)=0;
		virtual Matrix &operator*= (int c)=0;
		virtual Matrix &operator*= (double c)=0;		
		virtual Matrix &operator*= (MatrixC &op)=0;
		virtual Matrix &operator*= (MatrixI &op)=0;
		virtual Matrix &operator*= (MatrixF &op)=0;

		virtual Matrix &operator/= (unsigned char c)=0;
		virtual Matrix &operator/= (int c)=0;
		virtual Matrix &operator/= (double c)=0;		
		virtual Matrix &operator/= (MatrixC &op)=0;
		virtual Matrix &operator/= (MatrixI &op)=0;
		virtual Matrix &operator/= (MatrixF &op)=0;

		virtual Matrix &Multiply (MatrixF &op)=0;
		virtual Matrix &Size (int r, int c)=0;
		virtual Matrix &SizeSafe (int r, int c)=0;
		virtual Matrix &InsertRow (int r)=0;
		virtual Matrix &InsertCol (int c)=0;
		virtual Matrix &Transpose (void)=0;
		virtual Matrix &Identity (int order)=0;
		/*inline Matrix &RotateX (double ang);
		inline Matrix &RotateY (double ang);
		inline Matrix &RotateZ (double ang); */
		virtual Matrix &Basis (Vector3DF &c1, Vector3DF &c2, Vector3DF &c3)=0;
		virtual Matrix &GaussJordan (MatrixF &b)=0;

		virtual int GetRows(void)=0;
		virtual int GetCols(void)=0;
		virtual int GetLength(void)=0;		

		virtual unsigned char *GetDataC (void)=0;
		virtual int	*GetDataI (void)=0;
		virtual double *GetDataF (void)=0;
	};
	
	// MatrixC Declaration	
	#define VNAME		C
	#define VTYPE		unsigned char

	class MatrixC {
	public:
		VTYPE *data;
		int rows, cols, len;		

		// Constructors/Destructors
		inline MatrixC ();
		inline ~MatrixC ();
		inline MatrixC (int r, int c);

		// Member Functions
		inline VTYPE &operator () (int r, int c);
		inline MatrixC &operator= (unsigned char c);
		inline MatrixC &operator= (int c);
		inline MatrixC &operator= (double c);		
		inline MatrixC &operator= (MatrixC &op);
		inline MatrixC &operator= (MatrixI &op);
		inline MatrixC &operator= (MatrixF &op);
		
		inline MatrixC &operator+= (unsigned char c);
		inline MatrixC &operator+= (int c);
		inline MatrixC &operator+= (double c);		
		inline MatrixC &operator+= (MatrixC &op);
		inline MatrixC &operator+= (MatrixI &op);
		inline MatrixC &operator+= (MatrixF &op);

		inline MatrixC &operator-= (unsigned char c);
		inline MatrixC &operator-= (int c);
		inline MatrixC &operator-= (double c);		
		inline MatrixC &operator-= (MatrixC &op);
		inline MatrixC &operator-= (MatrixI &op);
		inline MatrixC &operator-= (MatrixF &op);

		inline MatrixC &operator*= (unsigned char c);
		inline MatrixC &operator*= (int c);
		inline MatrixC &operator*= (double c);		
		inline MatrixC &operator*= (MatrixC &op);
		inline MatrixC &operator*= (MatrixI &op);
		inline MatrixC &operator*= (MatrixF &op);

		inline MatrixC &operator/= (unsigned char c);
		inline MatrixC &operator/= (int c);
		inline MatrixC &operator/= (double c);		
		inline MatrixC &operator/= (MatrixC &op);
		inline MatrixC &operator/= (MatrixI &op);
		inline MatrixC &operator/= (MatrixF &op);

		inline MatrixC &Multiply (MatrixF &op);
		inline MatrixC &Size (int r, int c);
		inline MatrixC &SizeSafe (int r, int c);
		inline MatrixC &InsertRow (int r);
		inline MatrixC &InsertCol (int c);
		inline MatrixC &Transpose (void);
		inline MatrixC &Identity (int order);		
		inline MatrixC &Basis (Vector3DF &c1, Vector3DF &c2, Vector3DF &c3);
		inline MatrixC &GaussJordan (MatrixF &b);

		inline int GetRows(void);
		inline int GetCols(void);
		inline int GetLength(void);
		inline VTYPE *GetData(void);

		inline unsigned char *GetDataC (void)	{return data;}
		inline int *GetDataI (void)				{return NULL;}
		inline double *GetDataF (void)			{return NULL;}		
	};
	#undef VNAME
	#undef VTYPE

	// MatrixI Declaration	
	#define VNAME		I
	#define VTYPE		int

	class MatrixI {
	public:
		VTYPE *data;
		int rows, cols, len;		
	
		// Constructors/Destructors
		inline MatrixI ();
		inline ~MatrixI ();
		inline MatrixI (int r, int c);

		// Member Functions
		inline VTYPE &operator () (int r, int c);
		inline MatrixI &operator= (unsigned char c);
		inline MatrixI &operator= (int c);
		inline MatrixI &operator= (double c);		
		inline MatrixI &operator= (MatrixC &op);
		inline MatrixI &operator= (MatrixI &op);
		inline MatrixI &operator= (MatrixF &op);
		
		inline MatrixI &operator+= (unsigned char c);
		inline MatrixI &operator+= (int c);
		inline MatrixI &operator+= (double c);		
		inline MatrixI &operator+= (MatrixC &op);
		inline MatrixI &operator+= (MatrixI &op);
		inline MatrixI &operator+= (MatrixF &op);

		inline MatrixI &operator-= (unsigned char c);
		inline MatrixI &operator-= (int c);
		inline MatrixI &operator-= (double c);		
		inline MatrixI &operator-= (MatrixC &op);
		inline MatrixI &operator-= (MatrixI &op);
		inline MatrixI &operator-= (MatrixF &op);

		inline MatrixI &operator*= (unsigned char c);
		inline MatrixI &operator*= (int c);
		inline MatrixI &operator*= (double c);		
		inline MatrixI &operator*= (MatrixC &op);
		inline MatrixI &operator*= (MatrixI &op);
		inline MatrixI &operator*= (MatrixF &op);

		inline MatrixI &operator/= (unsigned char c);
		inline MatrixI &operator/= (int c);
		inline MatrixI &operator/= (double c);		
		inline MatrixI &operator/= (MatrixC &op);
		inline MatrixI &operator/= (MatrixI &op);
		inline MatrixI &operator/= (MatrixF &op);

		inline MatrixI &Multiply (MatrixF &op);
		inline MatrixI &Size (int r, int c);
		inline MatrixI &SizeSafe (int r, int c);
		inline MatrixI &InsertRow (int r);
		inline MatrixI &InsertCol (int c);
		inline MatrixI &Transpose (void);
		inline MatrixI &Identity (int order);		
		inline MatrixI &Basis (Vector3DF &c1, Vector3DF &c2, Vector3DF &c3);
		inline MatrixI &GaussJordan (MatrixF &b);

		inline int GetRows(void);
		inline int GetCols(void);
		inline int GetLength(void);
		inline VTYPE *GetData(void);

		inline unsigned char *GetDataC (void)	{return NULL;}
		inline int *GetDataI (void)				{return data;}
		inline double *GetDataF (void)			{return NULL;}
	};
	#undef VNAME
	#undef VTYPE

	// MatrixF Declaration	
	#define VNAME		F
	#define VTYPE		double

	class MatrixF {
	public:	
		VTYPE *data;
		int rows, cols, len;		

		// Constructors/Destructors		
		inline MatrixF ();
		inline ~MatrixF ();
		inline MatrixF (int r, int c);

		// Member Functions
		inline VTYPE &operator () (int r, int c);
		inline MatrixF &operator= (unsigned char c);
		inline MatrixF &operator= (int c);
		inline MatrixF &operator= (double c);		
		inline MatrixF &operator= (MatrixC &op);
		inline MatrixF &operator= (MatrixI &op);
		inline MatrixF &operator= (MatrixF &op);
		
		inline MatrixF &operator+= (unsigned char c);
		inline MatrixF &operator+= (int c);
		inline MatrixF &operator+= (double c);		
		inline MatrixF &operator+= (MatrixC &op);
		inline MatrixF &operator+= (MatrixI &op);
		inline MatrixF &operator+= (MatrixF &op);

		inline MatrixF &operator-= (unsigned char c);
		inline MatrixF &operator-= (int c);
		inline MatrixF &operator-= (double c);		
		inline MatrixF &operator-= (MatrixC &op);
		inline MatrixF &operator-= (MatrixI &op);
		inline MatrixF &operator-= (MatrixF &op);

		inline MatrixF &operator*= (unsigned char c);
		inline MatrixF &operator*= (int c);
		inline MatrixF &operator*= (double c);		
		inline MatrixF &operator*= (MatrixC &op);
		inline MatrixF &operator*= (MatrixI &op);
		inline MatrixF &operator*= (MatrixF &op);		

		inline MatrixF &operator/= (unsigned char c);
		inline MatrixF &operator/= (int c);
		inline MatrixF &operator/= (double c);		
		inline MatrixF &operator/= (MatrixC &op);
		inline MatrixF &operator/= (MatrixI &op);
		inline MatrixF &operator/= (MatrixF &op);

		inline MatrixF &Multiply4x4 (MatrixF &op);
		inline MatrixF &Multiply (MatrixF &op);
		inline MatrixF &Size (int r, int c);
		inline MatrixF &SizeSafe (int r, int c);
		inline MatrixF &InsertRow (int r);
		inline MatrixF &InsertCol (int c);
		inline MatrixF &Transpose (void);
		inline MatrixF &Identity (int order);
		inline MatrixF &RotateX (double ang);
		inline MatrixF &RotateY (double ang);
		inline MatrixF &RotateZ (double ang);
		inline MatrixF &Basis (Vector3DF &c1, Vector3DF &c2, Vector3DF &c3);
		inline MatrixF &GaussJordan (MatrixF &b);

		inline int GetRows(void);
		inline int GetCols(void);
		inline int GetLength(void);
		inline VTYPE *GetData(void);
		inline void GetRowVec (int r, Vector3DF &v);

		inline unsigned char *GetDataC (void)	{return NULL;}
		inline int *GetDataI (void)				{return NULL;}
		inline double *GetDataF (void)			{return data;}
	};
	#undef VNAME
	#undef VTYPE

    // Matrix Code Definitions (Inlined)

	#include "gamex-matrix.cci"

#endif